using System.Net;
using Common;
using Microsoft.AspNetCore.Mvc;

namespace Router
{
    struct Response
    {
        public bool IsSuccess;
        public object? Data;
    }

    /// <summary>
    /// Web API 
    /// </summary>
    class API
    {
        /// <summary>
        /// 发送字符串
        /// </summary>
        /// <param name="address">IPV4 或者 IPV6 地址</param>
        /// <param name="port">设备端口号</param>
        /// <param name="text">Text for send</param>
        /// <returns>Json: true or false</returns>
        public static async ValueTask<IResult> SendString(string address, int port, string text)
        {
            var endPoint = new IPEndPoint(IPAddress.Parse(address), port);
            var ret = await UDPClientPool.SendStringAsync(endPoint, [text]);
            if (ret) { return Results.Json(new Response() { IsSuccess = true }); }
            else { return Results.Json(new Response() { IsSuccess = false }); }
        }


        [HttpPost]
        public static async ValueTask<IResult> SendAddrPackage(
            string address,
            int port,
            // WebProtocol.BurstType burstType,
            // byte commandID,
            // bool isWrite,
            // byte burstLength,
            // UInt32 devAddress)
            WebProtocol.SendAddrPackOptions opts)
        {
            // WebProtocol.SendAddrPackOptions opts;
            // opts.burstType = burstType;
            // opts.commandID = commandID;
            // opts.isWrite = isWrite;
            // opts.burstLength = burstLength;
            // opts.address = devAddress;

            var endPoint = new IPEndPoint(IPAddress.Parse(address), port);
            var ret = await UDPClientPool.SendAddrPackAsync(endPoint, new WebProtocol.SendAddrPackage(opts));
            if (ret) { return Results.Json(new Response() { IsSuccess = true }); }
            else { return Results.Json(new Response() { IsSuccess = false }); }
        }

        public static async ValueTask<IResult> SendDataPackage(string address, int port, string data)
        {
            var endPoint = new IPEndPoint(IPAddress.Parse(address), port);
            var ret = await UDPClientPool.SendDataPackAsync(endPoint,
                new WebProtocol.SendDataPackage(NumberProcessor.StringToBytes(data).Value));
            if (ret) { return Results.Json(new Response() { IsSuccess = true }); }
            else { return Results.Json(new Response() { IsSuccess = false }); }
        }

        public class Jtag
        {
            /// <summary>
            /// 执行一个Jtag命令
            /// </summary>
            /// <param name="address"> 设备地址 </param>
            /// <param name="port"> 设备端口 </param>
            /// <param name="hexDevAddr"> 16进制设备目的地址(Jtag) </param>
            /// <param name="hexCmd"> 16进制命令 </param>
            /// <param name="hexExRet"> 16进制预期结果 </param>
            /// <returns> Response </returns>
            public static async ValueTask<IResult> RunCommand(string address, int port, string hexDevAddr, string hexCmd, string hexExRet)
            {
                var jtagCtrl = new JtagController.Jtag(address, port);
                var ret = await jtagCtrl.RunCommand(Convert.ToUInt32(hexDevAddr, 16), Convert.ToUInt32(hexCmd, 16), Convert.ToUInt32(hexExRet, 16));


                if (ret.IsSuccessful) { return Results.Json(new Response() { IsSuccess = true }); }
                else { return Results.Json(new Response() { IsSuccess = false, Data = ret.Error }); }
            }


            /// <summary>
            /// 获取Jtag ID Code
            /// </summary>
            /// <param name="address"> 设备地址 </param>
            /// <param name="port"> 设备端口 </param>
            /// <returns> Response </returns>
            public static async ValueTask<IResult> GetDeviceIDCode(string address, int port)
            {
                var jtagCtrl = new JtagController.Jtag(address, port);
                var ret = await jtagCtrl.ReadIDCode();

                if (ret.IsSuccessful) { return Results.Json(new Response() { IsSuccess = true, Data = ret.Value }); }
                else { return Results.Json(new Response() { IsSuccess = false, Data = ret.Error }); }
            }


        }
    }
}
