using DotNext;
using Newtonsoft.Json;

namespace WebProtocol
{

    /// <summary> The Sign of Package </summary>
    public enum PackSign
    {
        /// <summary> Package: Send Read or Write Address to Device </summary>
        SendAddr = 0x00,
        /// <summary> Package: Send Data Which Update the flash of Device</summary>
        SendData = 0xFF,
        /// <summary>  </summary>
        RecvData = 0x0F,
        /// <summary>  </summary>
        RecvResp = 0xF0,
    }

    /// <summary> Package Burst Type </summary>
    public enum BurstType
    {
        /// <summary> Extended Type </summary>
        ExtendBurst = 0b00,
        /// <summary> Fixed Type </summary>
        FixedBurst = 0b01,
    }


    /// <summary> Package options which to send address to read or write </summary>
    /// <example> 
    /// {
    ///    "burType":0,
    ///    "commandID":1,
    /// }
    /// </example>
    public struct SendAddrPackOptions
    {

        /// <summary> Package Burst Type </summary>
        [JsonProperty("burstType")]
        public BurstType burstType;
        /// <example>1</example>
        public byte commandID;
        /// <summary> 标识写入还是读取 </summary>
        /// <example> true </example>
        public bool isWrite;
        /// <summary> 突发长度：0是32bits，255是32bits x 256 </summary>
        /// <example> 255 </example>
        public byte burstLength;
        /// <example> 0x10_00_00_00 </example>
        public UInt32 address;

        /// <summary>
        /// Convert to Json String
        /// </summary>
        /// <returns> Json String </returns>
        public override string ToString()
        {
            return JsonConvert.SerializeObject(this);
        }
    }

    /// <summary> Package Options which to receive from boards </summary>
    public struct RecvPackOptions
    {
        /// <summary> 数据包类型 </summary>
        public enum PackType
        {
            /// <summary> 读响应包 </summary>
            ReadResp,
            /// <summary> 写响应包 </summary>
            WriteResp
        };

        /// <summary>
        /// 数据包类型
        /// </summary>
        public PackType type;
        /// <summary>
        /// Task ID
        /// </summary>
        public byte commandID;
        /// <summary>
        /// Whether is succeed to finish command
        /// </summary>
        public bool isSuccess;
        /// <summary>
        /// Return Data
        /// </summary>
        public byte[]? data;

        /// <summary>
        /// Convert to Json String
        /// </summary>
        /// <returns> Json String </returns>
        public override string ToString()
        {
            return JsonConvert.SerializeObject(this);
        }
    }

    /// <summary> Server->FPGA 地址包 </summary>
    public struct SendAddrPackage
    {
        readonly byte sign = (byte)PackSign.SendAddr;
        readonly byte commandType;
        readonly byte burstLength;
        readonly byte _reserved = 0;
        readonly UInt32 address;

        /// <summary>
        /// 使用地址包选项构造地址包
        /// </summary>
        /// <param name="opts"> 地址包选项 </param>
        public SendAddrPackage(SendAddrPackOptions opts)
        {
            byte byteBurstType = Convert.ToByte((byte)opts.burstType << 6);
            byte byteCommandID = Convert.ToByte((opts.commandID & 0x03) << 4);
            byte byteIsWrite = (opts.isWrite ? (byte)0x01 : (byte)0x00);
            this.commandType = Convert.ToByte(byteBurstType | byteCommandID | byteIsWrite);
            this.burstLength = opts.burstLength;
            this.address = opts.address;
        }

        /// <summary>
        /// 使用完整参数构造地址包
        /// </summary>
        /// <param name="burstType"> 突发类型 </param>
        /// <param name="commandID"> 任务ID </param>
        /// <param name="isWrite"> 是否是写数据 </param>
        /// <param name="burstLength"> 突发长度 </param>
        /// <param name="address"> 设备地址 </param>
        public SendAddrPackage(BurstType burstType, byte commandID, bool isWrite, byte burstLength, UInt32 address)
        {
            byte byteBurstType = Convert.ToByte((byte)burstType << 6);
            byte byteCommandID = Convert.ToByte((commandID & 0x03) << 4);
            byte byteIsWrite = (isWrite ? (byte)0x01 : (byte)0x00);
            this.commandType = Convert.ToByte(byteBurstType | byteCommandID | byteIsWrite);
            this.burstLength = burstLength;
            this.address = address;
        }

        public SendAddrPackage(byte commandType, byte burstLength, UInt32 address)
        {
            this.commandType = commandType;
            this.burstLength = burstLength;
            this.address = address;
        }

        /// <summary>
        /// 将地址包转化为字节数组
        /// </summary>
        /// <returns> 字节数组 </returns>
        public byte[] ToBytes()
        {
            var arr = new byte[8];
            arr[0] = sign;
            arr[1] = commandType;
            arr[2] = burstLength;
            arr[3] = _reserved;

            var bytesAddr = Common.NumberProcessor.NumberToBytes(address, 4).Value;
            Array.Copy(bytesAddr, 0, arr, 4, bytesAddr.Length);

            return arr;
        }

        /// <summary>
        /// 讲地址包转化为Json格式的地址包选项
        /// </summary>
        /// <returns> 字符串 </returns>
        public override string ToString()
        {
            SendAddrPackOptions opts;
            opts.burstType = (BurstType)(commandType >> 6);
            opts.commandID = Convert.ToByte((commandType >> 4) & 0b0011);
            opts.isWrite = Convert.ToBoolean(commandType & 0x01);
            opts.burstLength = burstLength;
            opts.address = address;

            return JsonConvert.SerializeObject(opts);
        }

        public static Result<SendAddrPackage> FromBytes(byte[] bytes, bool checkSign = true)
        {
            if (bytes.Length != 8)
            {
                throw new ArgumentException(
                    "Bytes are not equal to 8 bytes.",
                    nameof(bytes)
                );
            }

            if (checkSign && bytes[0] != (byte)PackSign.SendAddr)
            {
                throw new ArgumentException(
                    "The sign of bytes is not SendAddr Package, but you can disable it by set checkSign false.",
                    nameof(bytes)
                );
            }

            var address = Common.NumberProcessor.BytesToNumber(bytes[4..]).Value;
            return new SendAddrPackage(bytes[1], bytes[2], Convert.ToUInt32(address));
        }
    }

    public struct SendDataPackage
    {
        readonly byte sign = (byte)PackSign.SendData;
        readonly byte[] _reserved = new byte[3];
        readonly byte[] bodyData;

        public SendDataPackage(byte[] bodyData)
        {
            this.bodyData = bodyData;
        }

        public byte[] ToBytes()
        {
            var bodyDataLen = bodyData.Length;
            var arr = new byte[4 + bodyDataLen];

            arr[0] = sign;

            Array.Copy(bodyData, 0, arr, 4, bodyDataLen);

            return arr;
        }


    }

    /// <summary> FPGA->Server 读数据包 </summary>
    public struct RecvDataPackage
    {
        readonly byte sign = (byte)PackSign.RecvData;
        readonly byte commandID;
        readonly byte resp;
        readonly byte _reserved = 0;
        readonly byte[] bodyData;

        /// <summary>
        ///  FPGA->Server 读数据包 
        ///  构造函数
        /// </summary>
        /// <param name="commandID"> 任务ID号 </param>
        /// <param name="resp"> 读数据包响应 </param>
        /// <param name="bodyData"> 数据 </param>
        public RecvDataPackage(byte commandID, byte resp, byte[] bodyData)
        {
            this.commandID = commandID;
            this.resp = resp;
            this.bodyData = bodyData;
        }

        /// <summary>
        /// 获取读数据包选项
        /// </summary>
        public RecvPackOptions Options
        {
            get
            {
                RecvPackOptions opts;
                opts.type = RecvPackOptions.PackType.ReadResp;
                opts.commandID = commandID;
                opts.isSuccess = Convert.ToBoolean((resp >> 1) == 0b01 ? false : true);
                opts.data = bodyData;

                return opts;
            }
        }

        public static Result<RecvDataPackage> FromBytes(byte[] bytes)
        {
            if (bytes[0] != (byte)PackSign.RecvData)
                throw new ArgumentException(
                    "The sign of bytes is not RecvData Package!",
                    nameof(bytes)
                );
            return new RecvDataPackage(bytes[1], bytes[2], bytes[4..]);
        }
    }

    public struct RecvRespPackage
    {
        readonly byte sign = (byte)PackSign.RecvResp;
        readonly byte commandID;
        readonly byte resp;
        readonly byte _reserved = 0;

        public RecvRespPackage(byte commandID, byte resp)
        {
            this.commandID = commandID;
            this.resp = resp;
        }

        public RecvPackOptions Options
        {
            get
            {
                RecvPackOptions opts;
                opts.type = RecvPackOptions.PackType.WriteResp;
                opts.commandID = commandID;
                opts.isSuccess = Convert.ToBoolean((resp >> 1) == 0b01 ? false : true);
                opts.data = null;

                return opts;
            }
        }

        public static Result<RecvRespPackage> FromBytes(byte[] bytes)
        {
            if (bytes[0] != (byte)PackSign.RecvResp)
                throw new ArgumentException(
                    "The sign of bytes is not RecvData Package!",
                    nameof(bytes)
                );
            return new RecvRespPackage(bytes[1], bytes[2]);
        }
    }
}
